#ifndef INIFILE_CPP
#define INIFILE_CPP

#include <map>
#include <string>
#include <fstream>
#include <iostream>
#include <sstream>

using namespace std;

class inifile {
 public:
  typedef int64_t int_t;
  typedef unsigned int setind_t;
 private:
  string filename;
  map<string, string>* strmap;
  map<string, int_t>* lngmap;
  map<string, double>* dblmap;
  mutable map<string, string>::const_iterator strit;
  mutable map<string, int_t>::const_iterator lngit;
  mutable map<string, double>::const_iterator dblit;
  void mergeinto(const setind_t from, const setind_t to);
 public:
  mutable short good;
  bool complain_default;
  short fileexists;
  setind_t sets;
  inifile(const string fname, bool read = true);
  ~inifile();
  string getstr(const string& key, const string& defaultval, const setind_t setn = 0) const;
  int_t getlng(const string& key, const int_t defaultval, const setind_t setn = 0) const;
  double getdbl(const string& key, const double defaultval, const setind_t setn = 0) const;
  void setstr(const string& key, const string& value, const setind_t setn = 0);
  void writeini(ostream& of, const string& indent, const setind_t setn = 0) const;
  void reread();
  void defcomplain(const string& key) const;
};



#define INISTART "#--#INIFILE START#--#"
#define INISTOP "#--#INIFILE STOP#--#"


// Good gibt an, wie die letzte getvalue Abfrage gelaufen ist:
// 1: Wert wurde aus der Datei erfolgreich ausgelesen und ggf. umgewandelt
// 0: Wert wurde nicht in Datei gefunden, default wurde zurueckgegeben
// -1: Umwandlung (string2num) hat einen Fehler ergeben (OBSOLETE)
// 2: Angegebenes set wurde nicht gefunden, default value zurueckgegeben

inifile::inifile(const string fname, bool read) 
  : filename(fname), fileexists(0), complain_default(false) {
  strmap = NULL;
  lngmap = NULL;
  dblmap = NULL;
  if (read) reread();
}

inifile::~inifile() {
    delete[] strmap;
    delete[] lngmap;
    delete[] dblmap;
}

void inifile::reread() {
  delete[] strmap;
  delete[] lngmap;
  delete[] dblmap;
  string zeile, key, value;
  ifstream tifile(filename.c_str(), ios::in);
  int trenner;
  sets = 1;
  setind_t cs;
  setind_t incls;
  if (tifile.fail()) {
    fileexists = 0;
    cerr << "Unable to open file " << filename << ", using default values." << endl;
  } else fileexists = 1;
  
  if (fileexists) {
    // count number of datasets
    while (!tifile.eof()) {
      getline(tifile, zeile);
      if (zeile[0] == '[') {
	sscanf(zeile.c_str(), "[%u]", &cs);
	if (cs > sets-1) sets++;
      }
    }
    tifile.close();
    ifstream ifile(filename.c_str(), ios::in);
    // generate maps
    strmap = new map<string, string>[sets];
    lngmap = new map<string, inifile::int_t>[sets];
    dblmap = new map<string, double>[sets];

    // fill the entries in the maps
    cs = 0;
    stringstream conv;
    while (!ifile.eof()) {
      getline(ifile, zeile);
      if (zeile[0] == '#') continue;
      if (zeile[0] == '[') {
	if (sscanf(zeile.c_str(), "[%u]", &cs) != 1) cerr << "Error in inifile syntax!" << endl;
	continue;
      }
      if ( (zeile[0] == '\\') && (sscanf(zeile.c_str(), "\\%u", &incls) == 1) && (incls < sets) ) {
	mergeinto(incls, cs);
	continue;
      }

      trenner = zeile.find("=");
      key = zeile.substr(0,trenner);
      if (zeile.length() > trenner+1) {
	value = zeile.substr(trenner+1);
	// add to strmap, no matter what type
	strmap[cs][key] = value;
	// determine of what type is value?
	if ( ((value[0] >= '0') && (value[0] <= '9')) || (value[0] == '.') || (value[0] == '-' ) ) {  // it's a number => enter also in lngmap or dblmap
	  conv.clear();
	  conv.str(value);
	  if ( (value.find_first_of(".eE") == string::npos) )  // it's a long
	    conv >> lngmap[cs][key];
	  else                                    // it's a double
	    conv >> dblmap[cs][key];
	}
      } 
    }
    ifile.close();
  }
}

void inifile::mergeinto(const setind_t from, const setind_t to) {
  if ( (from >= sets) || (to >= sets) || !fileexists ) return;
  strmap[to].insert(strmap[from].begin(), strmap[from].end());
  lngmap[to].insert(lngmap[from].begin(), lngmap[from].end());
  dblmap[to].insert(dblmap[from].begin(), dblmap[from].end());
}

string inifile::getstr(const string& key, const string& defaultval, const setind_t setn) const {
  if (!fileexists) {
    good = 0;
    if (complain_default) defcomplain(key);
    return defaultval;
  }
  if (setn > sets-1) {
    good = 2;
    if (complain_default) defcomplain(key);
    return defaultval;
  }
  strit = strmap[setn].find(key);
  if (strit == strmap[setn].end()) {
    good = 0;
    if (complain_default) defcomplain(key);
    return defaultval;
  }
  else {
    good = 1;
    return strit->second;
  }
}


inifile::int_t inifile::getlng(const string& key, const inifile::int_t defaultval, const setind_t setn) const {
  if (!fileexists) {
    good = 0;
    if (complain_default) defcomplain(key);
    return defaultval;
  }
  if (setn > sets-1) {
    good = 2;
    if (complain_default) defcomplain(key);
    return defaultval;
  }
  lngit = lngmap[setn].find(key);
  if (lngit == lngmap[setn].end()) { 
    good = 0;
    if (complain_default) defcomplain(key);
    return defaultval;
  }
  else {
    good = 1;
    return lngit->second;
  }
}

double inifile::getdbl(const string& key, const double defaultval, const setind_t setn) const {
  if (!fileexists) {
    good = 0;
    if (complain_default) defcomplain(key);
    return defaultval;
  }
  if (setn > sets-1) {
    good = 2;
    if (complain_default) defcomplain(key);
    return defaultval;
  }
  dblit = dblmap[setn].find(key);
  if (dblit == dblmap[setn].end()) {
    good = 0;
    if (complain_default) defcomplain(key);
    return defaultval;
  }
  else {
    good = 1;
    return dblit->second;
  }
}


void inifile::setstr(const string& key, const string& value, const setind_t setn) {
  if (!fileexists) {
    good = 0;
    return;
  }
  if (setn > sets-1) {
    good = 2;
    return;
  }
  strmap[setn][key] = value;
}

void inifile::writeini(ostream& of, const string& indent, const setind_t setn) const {
  if (!fileexists) {
    good = 0;
    return;
  }
  setind_t s = setn;
  if (setn > sets-1) s = 0;
  of << indent << INISTART << endl;
  if (sets > 1) of << indent << '[' << s << ']' << endl;
  map<string, string>::const_iterator it = strmap[s].begin();
  for (it = strmap[s].begin(); it != strmap[s].end(); it++) 
    of << indent << it->first << "=" << it->second << endl;
  of << indent << INISTOP << endl;
}

void inifile::defcomplain(const string& key) const {
  std::cerr << "inifile: WARNING, cannot find key " << key << ", using default value." << endl;
}

int extractini(const string& sourcefn, const string& inifilefn) {
  int lines = 0;
  ifstream inf(sourcefn.c_str(), ios::in);
  ofstream outf(inifilefn.c_str(), ios::out);
  if (!inf) return -1;
  if (!outf) return -2;
  string zeile;
  unsigned long where;
  bool goon = true;
  do { 
    getline(inf, zeile);
    where = zeile.find(INISTART, 0);
  } while (!inf.eof() && (where == string::npos));
  while (!inf.eof() && goon) {
    getline(inf, zeile);
    if (zeile.find(INISTOP, 0) == string::npos) {
      outf << zeile.substr(where) << endl;
      lines++;
    } 
    else goon = false;
  } 
  inf.close();
  outf.close();
  return lines;
}

string extractstr(const string& sourcefn, const string& indent, const string& key, const string& defaultval) {
  string s, k;
  k = indent;
  k.append(key);
  k.append("=");
  bool notfound = true;
  ifstream inf(sourcefn.c_str(), ios::in);
  while (!inf.eof() && notfound) {
    getline(inf, s);
    if (s.substr(0,k.length()) == k) notfound = false;
  }
  if (notfound) {
    return defaultval;
  }
  else return s.substr(k.length());
}

      
#endif // INIFILE_CPP
